import os
import sys
import time

from abus import crypto
from abus import database

def list_archive(config):
   with database.open(config.database_path) as db:
      results= db.get_archive_contents(patterns=config.patterns,
                                       cutoff_date=config.cutoff,
                                       show_all=config.list_all)
      for path, timestamp, archive_dir, checksum in results:
         time_str= time.strftime("%Y-%m-%d %H:%M", time.localtime(timestamp))
         print("{}  {}".format(time_str, path))

def restore(config):
   with database.open(config.database_path) as db:
      results= db.get_archive_contents(patterns=config.patterns,
                                       cutoff_date=config.cutoff,
                                       show_all=config.list_all)
      results= list(results) # path, timestamp, archive_dir, checksum
      common= os.path.commonpath(ptac[0] for ptac in results)
      for path, timestamp, archive_dir, checksum in results:
         target= os.path.relpath(path, start=common)
         if config.list_all:
            sansext, ext= target= os.path.splitext(target)
            target= sansext +time.strftime("-%Y%m%d-%H%M", time.localtime(timestamp)) +ext
         if os.path.exists(target):
            print("Cannot overwrite", target, file=sys.stderr)
            continue

         parent= os.path.dirname(target)
         if parent=="" or os.path.isdir(parent):
            pass # good
         elif os.path.exists(parent):
            print("Cannot overwrite {} for {}".format(parent,target), file=sys.stderr)
            continue
         else:
            os.makedirs(parent)
         source= config.archive_root_path +"/" +archive_dir +"/" +checksum
         if os.path.exists(source+".z"):
            srcfd= crypto.open_lzma(source+".z", "r", config.password)
         else:
            srcfd= crypto.open_bin(source, "r", config.password)
         with srcfd as src, open(target, "wb") as dstfd:
            while True:
               blk= src.read(8192)
               if len(blk)==0: break
               dstfd.write(blk)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import base64
import lzma
import tarfile
import os
# pip install cryptography
from cryptography.hazmat.primitives import ciphers, hmac, hashes
from cryptography.hazmat.primitives.kdf import pbkdf2
from cryptography.hazmat import backends

def _password_derive(key, salt):
   kdf = pbkdf2.PBKDF2HMAC(
      algorithm=hashes.SHA256(),
      length=16,
      salt=salt,
      iterations=100000,
      backend=backends.default_backend()
   )
   return base64.urlsafe_b64encode(kdf.derive(key))

class _WriteableEncryptedStream(object):
   """A write-only stream to an encrypted file. Data written to the stream is automatically encrypted."""
   def __init__(self, path, password):
      backend = backends.default_backend()
      str_key = password.encode()
      salt = os.urandom(16)
      hmac_salt = os.urandom(16)
      iv = os.urandom(16)
      key = _password_derive(str_key, salt)
      hmac_key = _password_derive(str_key, hmac_salt)
      self.checksummer = hmac.HMAC(hmac_key, hashes.SHA256(), backend=backend)
      cipher = ciphers.Cipher(ciphers.algorithms.AES(key), ciphers.modes.CTR(iv), backend=backend)
      self.cryptor = cipher.encryptor()
      initialiser= iv + salt + hmac_salt
      self.checksummer.update(initialiser)
      self.output_stream= open(path, mode="wb")
      self.output_stream.write(initialiser)
   def write(self, data):
      out = self.cryptor.update(data)
      self.checksummer.update(out)
      self.output_stream.write(out)
   def close(self):
      out = self.cryptor.finalize()
      self.checksummer.update(out)
      self.output_stream.write(out)
      signature = self.checksummer.finalize()
      self.output_stream.write(signature)
      self.output_stream.close()

class _ReadableEncryptedStream(object):
   """A read-only stream to an encrypted file. Data read from the stream is automatically decrypted."""
   def __init__(self, path, password):
      self.input_stream= open(path, "rb")
      str_key = password.encode()
      raw= self.input_stream.read(80)
      if len(raw)<80:
         raise RuntimeError("file too short")
      assert len(raw)==80
      initialiser= raw[:48]
      iv = initialiser[:16]
      salt = initialiser[16:32]
      hmac_salt = initialiser[32:48]
      self.read_ahead= raw[48:] # empty indicates end of input_stream
      self.outbuffer= b""
      key = _password_derive(str_key, salt)
      hmac_key = _password_derive(str_key, hmac_salt)
      backend = backends.default_backend()
      cipher = ciphers.Cipher(ciphers.algorithms.AES(key), ciphers.modes.CTR(iv), backend=backend)
      self.cryptor = cipher.decryptor()
      self.checksummer = hmac.HMAC(hmac_key, hashes.SHA256(), backend=backend)
      self.checksummer.update(initialiser)
   def _fill(self, size):
      if not self.read_ahead:
         return
      bs= 8192 if size<0 else size
      while size==-1 or size>len(self.outbuffer):
         raw= self.read_ahead + self.input_stream.read(bs)
         encrypted= raw[:-32]
         self.read_ahead= raw[-32:]
         if len(encrypted) > 0:
            self.outbuffer += self.cryptor.update(encrypted)
            self.checksummer.update(encrypted)
         else:
            # reached the end which means read_ahead is signature
            self.outbuffer += self.cryptor.finalize()
            self.checksummer.verify(self.read_ahead)
            self.read_ahead= None
            break
   def read(self, size=-1):
      self._fill(size)
      n= len(self.outbuffer) if size<0 or len(self.outbuffer)<size else size
      data= self.outbuffer[:size]
      self.outbuffer= self.outbuffer[size:]
      return data
   def seekable(self):
      return False
   def close(self):
      pass

class open_bin(object):
   """Opens an encrypted binary stream in a context manager."""
   def __init__(self, path, mode, password):
      if mode == "w":
         fd= _WriteableEncryptedStream(path, password)
      elif mode == "r":
         fd= _ReadableEncryptedStream(path, password)
      else:
         raise ValueError("Invalid mode: "+mode)
      # Derived classes may append more streams using fd as underlying stream.
      self._streams= [fd]
   def __enter__(self):
      return self._streams[-1]
   def __exit__(self, exc_type, exc_val, exc_tb):
      while self._streams:
         self._streams.pop().close()

class open_tar(open_bin):
   """Opens an encrypted and compressed tarfile.TarFile in a context manager."""
   def __init__(self, path, mode, password):
      open_bin.__init__(self, path, mode, password)
      tarfd= tarfile.open(mode=mode+"|xz", fileobj=self._streams[-1])
      self._streams.append(tarfd)

class open_txt(open_bin):
   """Opens an encrypted and compressed UTF-8 stream."""
   def __init__(self, path, mode, password):
      open_bin.__init__(self, path, mode, password)
      f= lzma.open(self._streams[-1], mode+"t", encoding="UTF-8")
      self._streams.append(f)

class open_lzma(open_bin):
   """Opens an encrypted and compressed UTF-8 stream."""
   def __init__(self, path, mode, password):
      open_bin.__init__(self, path, mode, password)
      f= lzma.open(self._streams[-1], mode+"b")
      self._streams.append(f)



import itertools
import os
import shutil
import unittest

from abus import database, main

class AbusTestBase(unittest.TestCase):
   @staticmethod
   def remove_dir_contents(path):
      for direntry in os.scandir(path):
         if direntry.is_dir():
            shutil.rmtree(direntry.path)
         elif direntry.is_file():
            os.unlink(direntry.path)

   @staticmethod
   def find_files(start):
      """Returns direntries for all files in `start`, recursively"""
      for direntry in os.scandir(start):
         if direntry.is_file():
            yield direntry
         elif direntry.is_dir():
            yield from AbusTestBase.find_files(direntry.path)

   @staticmethod
   def create_homedir_file(path, prime):
      os.makedirs(os.path.dirname(path), exist_ok=True)
      with open(path,"wb") as f:
         f.write(bytes(n*prime%256 for n in range(8192)))

   def setup_directories(self):
      candidates_for_tmp= ["C:/tmp", "C:/temp"]
      if "TEMP" in os.environ: candidates_for_tmp.append(os.environ["TEMP"])
      for tempdir in candidates_for_tmp:
         if os.path.isdir(tempdir):
            break
      else:
         raise Exception("Could not find a temporary directory for tests")

      self.password= "Sesam, öffne dich!"
      self.homedir= tempdir + "/abushome"
      self.archivedir= tempdir + "/abusarchive"
      self.restoredir= tempdir + "/abusrestore"
      os.makedirs(self.homedir, exist_ok=True)
      os.makedirs(self.archivedir, exist_ok=True)
      os.makedirs(self.restoredir, exist_ok=True)
      os.chdir(self.restoredir)

      # empty them
      self.remove_dir_contents(self.homedir)
      self.remove_dir_contents(self.archivedir)
      self.remove_dir_contents(self.restoredir)

      self.configfile= tempdir + "/abusconfig.cfg"
      with open(self.configfile, "w", encoding="utf-8") as f:
         print("logfile", tempdir+"/abuslog.txt", file=f)
         print("archive", self.archivedir, file=f)
         print("password", self.password, file=f)
         print("[include]", file=f)
         print(self.homedir, file=f)

      with database.open(self.databasepath, allow_create=True) as db:
         pass

   def setup_backup_with_well_known_checksums(self):
      self.setup_directories()
      self.expected_backups= set() # contains (path rel to homedir, backup filename) pairs

      path= "my_valueable_file"
      with open(self.homedir+"/"+path, "w", encoding="utf-8") as f:
         for n in range(400):
            for p in range(10):
               print(n**p, file=f, end="")
            print(file=f)
      self.expected_backups.add((path,"73672933c00ab3cd730c8715a392ee6dee9ba2c0a8e5e5d07170b6544b0113ef.z"))

      path= "my_little_file"
      with open(self.homedir+"/"+path, "wb") as f:
         f.write(bytearray(range(256)))
      self.expected_backups.add((path,"40aff2e9d2d8922e47afd4648e6967497158785fbd1da870e7110266bf944880"))

      path= "subdir_a/I_am_in_a.tif"
      self.create_homedir_file(self.homedir+"/"+path, 31)
      self.expected_backups.add((path,"498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"))

      path= "subdir_b/I_am_in_b.bin"
      self.create_homedir_file(self.homedir+"/"+path, 97)
      self.expected_backups.add((path,"1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3.z"))

      main.main(["test", "-f", self.configfile, "--backup"])

   def setup_multiple_backups(self):
      self.setup_directories()
      def primes():
         prev= [2]
         for c in itertools.count(3,2):
            for p in prev:
               if p*p>c:
                  yield c
                  prev.append(c)
                  break
               if c%p==0: break
      primes= primes().__iter__()

      all_paths= ["a", "b", "s/a", "s/b", "t/a", "t/b"]
      n= len(all_paths)*3//4
      for change_paths in all_paths, all_paths[:n], all_paths[-n:]:
         for path in change_paths:
            self.create_homedir_file(self.homedir+"/"+path, primes.__next__())
         main.main(["test", "-f", self.configfile, "--backup"])
         with database.open(self.databasepath) as db:
            db.testing_move_timestamps(-3000)
   @property
   def databasepath(self):
      return self.archivedir+"/index.sl3"



import unittest.mock
from abus import crypto, database, main
from abus.testbase import AbusTestBase

class BackedUpHomedirTests(AbusTestBase):
   def test_simple_end_to_end_backup(self):
      self.setup_backup_with_well_known_checksums()
      expect= {"index.sl3"}
      expect.update(s for p,s in self.expected_backups)
      result= set(direntry.name for direntry in self.find_files(self.archivedir))
      self.assertEqual(result, expect)

      # decrypting backup
      my_litte_backup= [e for e in self.find_files(self.archivedir) if e.name.startswith("40aff2e")]
      self.assertEqual(len(my_litte_backup), 1)
      self.assertEqual(my_litte_backup[0].stat().st_size, 256+80) # 80 for salt, init vector, and checksum
      with crypto.open_bin(my_litte_backup[0], "r", self.password) as f:
         blk= f.read(8192)
         self.assertEqual(len(blk), 256)
         self.assertEqual(blk, bytearray(range(256)))

   def test_error_prevents_closing_backup_run(self):
      self.setup_directories()
      self.create_homedir_file(self.homedir+"/unreadable",23)
      class mock_make_backup_copy(object):
         def __call__(self, path, expected_checksum, backup_path, open_dst_function, password):
            raise RuntimeError("file changed while reading: "+path)
      with unittest.mock.patch('abus.backup.make_backup_copy', new_callable=mock_make_backup_copy):
         main.main(["test", "-f", self.configfile, "--backup"])
      with database.open(self.databasepath) as db:
         with db._get_connection("test_error_prevents_closing_backup_run") as conn:
            rs= conn.execute("select timestamp from run").fetchall()
            self.assertEqual(rs, [(None,)])



import itertools
import sqlite3
from pprint import pprint
from abus import database
from abus.testbase import AbusTestBase

class DatabaseTests(AbusTestBase):
   def test_database_get_archive(self):
      self.setup_directories()
      checksums= ("e{:03}".format(i) for i in itertools.count()).__iter__()
      timestamps= itertools.count(1506159900, 50).__iter__()
      all_files= ["c:/home/file_{}.txt".format(c) for c in "abcdefghxy"]
      file_states= {}
      data= [] # (run_name, timestamp, path, checksum, archive_dir) list
      for run,files in [("2017-09-22 12:45","abcdefghxy"),
                        ("2017-09-23 12:45","xy"),
                        ("2017-09-24 12:45","bcdefghx"),
                        ("2017-09-25 12:45","abcfghy"),
                       ]:
         for c in files:
            file_states[c]= timestamps.__next__(), checksums.__next__()
         # this sometimes "touches" b, tests that some content is not listed twice:
         file_states["b"]= timestamps.__next__(), file_states["b"][1]
         data.extend((run,ts,"c:/home/file_{}.txt".format(c),cs,cs[:2]+"/"+cs[2:])
                     for c,(ts,cs) in file_states.items())
      with sqlite3.connect(self.databasepath) as conn:
         conn.executemany("insert into location(checksum, archive_dir) values(?,?)", set(t[3:5] for t in data))
         conn.executemany("insert into content(run_name, timestamp, path, checksum) values(?,?,?,?)", (t[0:4] for t in data))
         conn.execute("insert into run(run_name, timestamp) select run_name, max(timestamp) from content group by run_name")

      with database.open(self.databasepath) as db:
         def test_one(cutoff_date, show_all, nearly_expect):
            # nearly_expect may still contain duplicate (file,content) with different timestamps
            pacs= set((p,a,c)for p,t,a,c in nearly_expect)
            ptacs= [min(ptac for ptac in nearly_expect if ptac[0]==p and ptac[2]==a and ptac[3]==c) for p,a,c in pacs]
            expect= sorted(ptacs, key= lambda ptac: (ptac[0],-ptac[1]))
            result= list(db.get_archive_contents([], cutoff_date, show_all))
            self.assertEqual(result, expect)

            expect= [ptac for ptac in expect if ptac[0][-5] in "ab"]
            result= list(db.get_archive_contents(['*a.txt','*b.txt'], cutoff_date, show_all))
            self.assertEqual(result, expect)

            expect= [ptac for ptac in expect if ptac[0][-5]=="a"]
            result= list(db.get_archive_contents(['*a.txt'], cutoff_date, show_all))
            self.assertEqual(result, expect)

         test_one(cutoff_date=None, show_all=False,
                  nearly_expect= [(p,t,a,c) for r,t,p,c,a in data if r=="2017-09-25 12:45"],
                  )
         test_one(cutoff_date=None, show_all=True,
                  nearly_expect= [(p,t,a,c) for r,t,p,c,a in data],
                  )

         by_path= {}
         for cutoff in sorted(set(t for r,t,p,c,a in data)):
            try:
               test_one(cutoff_date=cutoff, show_all=True,
                        nearly_expect= [(p,t,a,c) for r,t,p,c,a in data if t<=cutoff],
                        )
               by_path.update({p:(p,t,a,c) for r,t,p,c,a in data if t==cutoff})
               test_one(cutoff_date=cutoff, show_all=False,
                        nearly_expect= by_path.values(),
                        )
            except:
               print("iteration data:", cutoff, by_path)
               raise



import logging
import os
import sys
import traceback

from abus import config
from abus import backup, database
from abus import list_restore

def main(argv):
   try:
      cfg= config.Configuration(argv)
      if cfg.is_backup:
         backup.do_backup(cfg)
      elif cfg.is_restore:
         list_restore.restore(cfg)
      elif cfg.is_init:
         if os.path.exists(cfg.database_path):
            raise config.ConfigurationError(cfg.database_path+" exists already")
         with database.open(cfg.database_path,allow_create=True):
            print("created", cfg.database_path)
      else:
         list_restore.list_archive(cfg)
   except config.ConfigurationError as e:
      error_message= str(e)
   except RuntimeError as e:
      error_message= str(e)
   except Exception as e:
      error_message= "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
   else:
      return 0

   print("Error:", error_message, file=sys.stderr)
   if logging.root.handlers: # tests that logging could be initialised before error occurred
      logging.error(error_message)

   return 1


#
# TODO test that errors are issued if directory cannot be read
#



import sys
import queue
import concurrent.futures
import threading
from contextlib import contextmanager

_data_locks= {}
_datalocks_lock= threading.Lock()
_lock_pool= [] # because creating locks must be expensive

@contextmanager
def global_data_lock(*data):
   """
   context manager for acquiring a lock specific to the given data
   :param data:
   :type data:
   :return:
   :rtype:
   """
   # A reference count is required with the lock so that it can be deleted when no-one is using it,
   # so that _data_locks does not grow indefinetely.
   with _datalocks_lock:
      if data in _data_locks:
         lock_and_count= _data_locks[data]
         lock_and_count[1] += 1
      else:
         lock= _lock_pool.pop() if _lock_pool else threading.Lock()
         lock_and_count= [lock, 1] # Using list instead of tuple because it needs to be mutable.
         _data_locks[data]= lock_and_count
   with lock_and_count[0]:
      yield
   with _datalocks_lock:
      if lock_and_count[1]==1:
         del _data_locks[data]
         _lock_pool.append(lock_and_count[0])
      else:
         lock_and_count[1] -= 1



import concurrent.futures
import enum
import hashlib
import logging
import os
import sys
from fnmatch import fnmatch

from abus import synchronisation
from abus import crypto
from abus import database

class Counters(enum.Enum):
   DIRS= "directories"
   UNTOUCHED= "untouched files"
   CHECKSUMMED_ONLY= "already backed up"
   COMPRESSED= "compressed"
   UNCOMPRESSED= "copied"
   ERROR= "errors"

def do_backup(config):
   with database.open(config.database_path) as db:
      run_name= db.open_backup_run()
      logging.info("starting backup run %s", run_name)
      iter_archivedirs_to_use= iter(archive_dirs_to_use(db))
      stats= {c:0 for c in Counters}
      with concurrent.futures.ThreadPoolExecutor() as executor:
         # This thread walks the directories, runs do_one_file in the background for each file
         # and then waits for all the background tasks.
         futures= []
         dir_queue= list(config.include)
         while dir_queue:
            p= dir_queue.pop()
            try:
               direntries= os.scandir(p)
            except PermissionError as e:
               logging.error("Error reading directory: %s", e)
               stats[Counters.ERROR] += 1
               continue
            else:
               stats[Counters.DIRS] += 1
            for direntry in direntries:
               path= direntry.path.replace('\\', '/')
               if any(fnmatch(path, p) for p in config.exclude):
                  pass
               elif direntry.is_file(follow_symlinks=False):
                  # restricts number of simultaneous single-file backups:
                  if len(futures)>40:
                     wait_for_one(futures, stats)
                  f= executor.submit(do_one_file, direntry, run_name, iter_archivedirs_to_use, config, db)
                  futures.append(f)
               elif direntry.is_dir(follow_symlinks=False):
                  dir_queue.append(path)
         while futures:
            wait_for_one(futures, stats)
      logging.info(", ".join("{}: {}".format(c.value, n) for c,n in stats.items()))
      n_errors= stats[Counters.ERROR]
      if n_errors==0:
         db.complete_backup_run(run_name)
         logging.info("completed backup run %s without errors", run_name)
      else:
         logging.warning("completed backup run %s with errors", run_name)

def wait_for_one(futures, stats):
   done,not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
   error_count= 0
   for f in done:
      try:
         stats[f.result()] += 1
      except Exception as e:
         logging.error(str(e))
         print(str(e), file=sys.stderr)
         stats[Counters.ERROR] += 1
   futures[:]= not_done

def archive_dirs_to_use(database):
   subdir_usage= {"{:02}/{:02}".format(i//100, i%100):0 for i in range(10000)}
   subdir_usage.update(database.getarchivedir_usage())
   for a,n in subdir_usage.items():
      for _ in range(n, 100):
         yield a
   while True:
      yield from subdir_usage

def calculate_checksum(path):
   partially_calculated_checksum= hashlib.sha256()
   with open(path, "rb") as fd:
      while True:
         blk= fd.read(8192)
         if len(blk)==0: break
         partially_calculated_checksum.update(blk)
   return partially_calculated_checksum.hexdigest()

def do_one_file(direntry, run_name, iter_archivedirs_to_use, config, database):
   """
   Processes a single file, making appropriate database entries for it and creating the
   backup copy if necessary. May raise excpeptions.

   :param direntry: a direntry with the file information as returned by os.scandir()
   :param run_name: the backup run to which the file should be added
   :param iter_archivedirs_to_use: iterator over archive dirs (relative to archive root) into which new backup copies
                                   should be placed.
   :param config: Config object
   :param database: Database object
   """
   result= Counters.UNTOUCHED
   path= direntry.path.replace("\\", "/")
   ext= os.path.splitext(path)[-1]
   stat= direntry.stat()
   timestamp= stat.st_mtime
   size= stat.st_size
   if stat.st_dev==0 and stat.st_ino==0:
      # this means this is windows and the direntry does not have the relevant information
      stat= os.stat(path)
   with synchronisation.global_data_lock(stat.st_dev, stat.st_ino, timestamp):
      checksum= database.lookup_checksum(stat.st_dev, stat.st_ino, timestamp)
      if checksum is None:
         checksum= calculate_checksum(path)
         database.remember_checksum(stat.st_dev, stat.st_ino, timestamp, checksum)
         result= Counters.CHECKSUMMED_ONLY

   database.add_backup_entry(run_name, path, timestamp, checksum)

   with synchronisation.global_data_lock(checksum):
      archive_dir_rel= database.lookup_archivedir(checksum)
      if archive_dir_rel is None:
         with synchronisation.global_data_lock():
            archive_dir_rel= next(iter_archivedirs_to_use)
         backup_path= config.backup_path_abs(archive_dir_rel, checksum)
         os.makedirs(os.path.dirname(backup_path), exist_ok=True)
         if size < config.minimum_size_for_compression or ext in config.compressed_extensions:
            make_backup_copy(path, checksum, backup_path, crypto.open_bin, config.password)
            result= Counters.UNCOMPRESSED
         else:
            backup_path += ".z"
            make_backup_copy(path, checksum, backup_path, crypto.open_lzma, config.password)
            result= Counters.COMPRESSED
         database.remember_archivedir(checksum, archive_dir_rel)
   return result

def make_backup_copy(path, expected_checksum, backup_path, open_dst_function, password):
   """
   Creates the encrypted backup copy of a file, taking care not to leave a partial backup file in case of errors.

   :param path: Path of file to be backed up.
   :param expected_checksum: Known checksum of the file, which is verified while copying and if different and exception
                             will be raised.
   :param backup_path: absolute filename of the destination file
   :param open_dst_function: function from crypto module to open the destination file, i.e. either open_lzma or open_bin according to whether
                             compression is required.
   :param password: to be used for encryption
   """
   partially_calculated_checksum= hashlib.sha256()
   try:
      with open(path,"rb") as src, open_dst_function(backup_path+".part", "w", password) as dst:
         while True:
            blk= src.read(8192)
            if len(blk)==0: break
            partially_calculated_checksum.update(blk)
            dst.write(blk)
      if partially_calculated_checksum.hexdigest() != expected_checksum:
         raise RuntimeError("file changed while reading: "+path)
   except:
      os.unlink(backup_path+".part")
      raise
   else:
      os.rename(backup_path+".part", backup_path)




import time
from abus import config
from abus.testbase import AbusTestBase

class ConfigTestser(AbusTestBase):
   def test_parse_date(self):
      def str2sec(s):
         return time.mktime(time.strptime(s, "%Y%m%d%H%M%S"))
      now= str2sec("20170912150000")
      for d in range(1,9):
         for t in range(8):
            for future_date in True, False:
               for future_time in True, False:
                  d_str= "20171003" if future_date else "20170912"
                  t_str= "-160723" if future_time else "-130905"
                  test_str= d_str[-d:] + t_str[:t]

                  if d in (4,6,8) and t in (0,5,7):
                     expect_str= "2016"+d_str[-4:] if future_date and d==4 else d_str
                     expect_str += (t_str[1:t]+"000000")[:6]
                     try:
                        self.assertEqual(config.parse_date(test_str, now), str2sec(expect_str))
                     except:
                        print("test string:", test_str)
                        raise
                  else:
                     with self.assertRaises(ValueError):
                        print("test string:", test_str, "result:", config.parse_date(test_str, now))




import os
from abus import main, backup, crypto, database
from abus.testbase import AbusTestBase

class RestoreTests(AbusTestBase):
   def test_simple_restore(self):
      self.setup_backup_with_well_known_checksums()
      self.remove_dir_contents(self.restoredir)
      main.main(["test", "-f", self.configfile, "-r"])
      self.assertEqual(sum(1 for _ in self.find_files(self.restoredir)), len(self.expected_backups))
      for relpath, archive_name in self.expected_backups:
         actual= backup.calculate_checksum(self.restoredir+"/"+relpath)
         expected= archive_name[:64]
         self.assertEqual(actual, expected)

   def test_restore_does_not_overwrite(self):
      self.setup_backup_with_well_known_checksums()
      self.remove_dir_contents(self.restoredir)
      filenames= [path.partition("/")[0] for path,_ in self.expected_backups]
      for filename in filenames:
         with open(filename,"w",newline='\n') as f: f.write("Hello World\n")
      main.main(["test", "-f", self.configfile, "-r"])
      for filename in filenames:
         actual= backup.calculate_checksum(filename)
         self.assertEqual(actual, "d2a84f4b8b650937ec8f73cd8be2c74add5a911ba64df27458ed8229da804a26")

   def test_no_partial_backupfile_is_left(self):
      self.setup_backup_with_well_known_checksums()
      relpath, archivename= max(self.expected_backups)
      path= self.homedir+"/"+relpath
      expected_checksum= "42" # causes Exception which is expected not to leave file behind
      password= "Would you like to buy some air?"
      open_dst_function= crypto.open_lzma
      backup_path= self.archivedir+"/42"
      with self.assertRaisesRegex(Exception, "file changed while reading"):
         backup.make_backup_copy(path, expected_checksum, backup_path, open_dst_function, password)
      self.assertFalse(os.path.exists(backup_path))
      self.assertFalse(os.path.exists(backup_path+".part"))
      # making doubly sure the file would normally have been created (the source path might just be wrong, for example):
      backup.make_backup_copy(path, archivename[:64], backup_path, open_dst_function, password)
      self.assertTrue(os.path.exists(backup_path))
      self.assertFalse(os.path.exists(backup_path+".part"))

   def test_allversions_restore(self):
      self.setup_multiple_backups()
      n_versions= sum(1 for direntry in self.find_files(self.archivedir)
                        if len(direntry.name) in(64,66))
      main.main(["test", "-f", self.configfile, "-r", "-a"])
      self.find_files(self.restoredir)
      n_restored= sum(1 for _ in self.find_files(self.restoredir))
      self.assertEqual(n_restored, n_versions)



from setuptools import setup

setup(name='abus',
      version='1',
      description='Abingdon Backup Script',
      author='Cornelius Grotjahn',
      author_email='s1@tempaddr.uk',
      license='private',
      packages=['abus'],
      install_requires=['cryptography'],
      zip_safe=False,
      entry_points = { 'console_scripts': ['abus=abus:entry_point'], }, # script=package:function
      )



import argparse
import logging
import os
import time

class ConfigurationError(Exception):
   pass

class Configuration(object):
   """Combined backup configuration options from command line, environment, and config file"""
   def __init__(self, argv):
      """
      Combined backup configuration options from command line, environment, and config file (determined
      by command line or environment.

      :param argv: command line arguments (`sys.argv`).
      """

      # Default values
      self.archive_root_path= None
      self.logfile_path= None
      self.include= []
      x= [
         "*#",
         "*.lck",
         "*.lock",
         "*.wav",
         "*/~*",
         "*/desktop.ini",
         "*/Ditto.db",
         "*/UsrClass.dat*",
         "*/ZbThumbnail.info",
         "*cache/*",
         "C:/cornie/.lesshst",
         "C:/cornie/.ssh",
         "C:/cornie/installers/cygwin",
         "C:/cornie/music",
         "C:/cornie/pictures",
         "C:/cornie/rawpictures",
         "C:/cornie/run/RYAplotter",
         "C:/cornie/scheduled.log",
         "C:/cornie/vimfiles",
         "C:/Users/*/Documents/My Music",
         "C:/Users/*/Documents/My Pictures",
         "C:/Users/*/Documents/My Videos",
         "C:/Users/philip/Documents/Need_For_Speed_Most_Wanted",
         "C:/Users/sonja/Documents/Trazenje posla",
         ]
      self.exclude= []
      self.password= None
      self.minimum_size_for_compression= 4096
      self.compressed_extensions= set([".jpg", ".tif", ".gz", ".tgz",])

      # Command line
      args= self._get_command_line_options(argv)
      if args.f is not None:
         config_path= args.f
      elif "ABUS_CONFIG" in os.environ:
         config_path= os.environ["ABUS_CONFIG"]
      else:
         raise ConfigurationError("Missing config file path (-f option or ABUS_CONFIG environment variable)")
      self.is_backup= args.backup
      self.is_restore= args.restore
      self.is_init= args.init
      self.list_all= args.a
      self.patterns= args.glob
      self.cutoff= args.d

      # Config file
      with open(config_path, encoding="utf-8") as f:
         try:
            self._parse_config_file(f)
         finally:
            # Initialising log file now because it might be possible despite a subsequent
            # error during parsing.
            if self.logfile_path:
               logging.basicConfig(filename=self.logfile_path,
                                   level=logging.DEBUG,
                                   format='%(asctime)s %(levelname)-7s %(message)s',
                                   datefmt='%Y-%m-%d %H:%M:%S',
                                   )

      if (self.is_init or self.is_backup) and (self.is_restore or self.list_all or self.cutoff) \
              or self.is_init and self.is_backup:
         raise ConfigurationError("conflicting command line options (--backup/--init/-r/-a/-d)")
      if not self.archive_root_path:
         raise ConfigurationError("archive directory option not set")
      if not os.path.isdir(self.archive_root_path):
         raise ConfigurationError("missing archive directory ({})".format(self.archive_root_path))
      if not self.password:
         raise ConfigurationError("password directory option not set")
      if not self.include:
         raise ConfigurationError("missing or empty [include] section")
      if not self.logfile_path:
         raise ConfigurationError("logfile path option not set")

   def _parse_config_file(self, lines):
      self.include= []
      self.exclude= []
      section= None
      for line in lines:
         line= line.strip()
         if not line or line.startswith('#'):
            pass
         elif line.lower().startswith("[incl"):
            section= self.include
         elif line.lower().startswith("[excl"):
            section= self.exclude
         elif line.lower().startswith("["):
            raise ConfigurationError("unknown section name: "+line)
         elif section is not None:
            line= line.replace("\\", "/")
            if line.endswith("/"): line= line[:-1]
            section.append(line)
         else:
            splut= line.split(maxsplit=1)
            if len(splut)!=2:
               raise ConfigurationError("missing value: "+line)
            keyword, args = splut
            if keyword=="archive":
               self.archive_root_path= args
            elif keyword=="logfile":
               self.logfile_path= args
            elif keyword=="password":
               self.password= args
            else:
               raise ConfigurationError("unknown option: "+keyword)

   def _get_command_line_options(self, argv):
      parser= argparse.ArgumentParser(prog=os.path.basename(argv[0]),
         #usage='PROG [-f configfile] --backup',
         description="The Abingdon BackUp Scripts can make copies of files "
         "(defined in the abus config file) to an archive location "
         "and restore the backups to the current directories. "
         "By default files in the archive are listed.",
         )
      parser.add_argument('-f',
         action='store',
         metavar='path',
         help="path to config file required for defining backup location etc. "
         "Defaults to value of the ABUS_CONFIG environment variable")
      parser.add_argument('--init',
                          action='store_true',
                          help='creates an empty database in archive directory')
      parser.add_argument('--backup',
         action='store_true',
         help="backs up files to archive rather than listing files in archive.")
      parser.add_argument('-r', '--restore',
         action='store_true',
         help="restores files from archive to current directory rather than listing files in archive.")
      parser.add_argument('-a',
         action='store_true',
         help="includes all matching file versions in listing or restore "
         "rather than only the latest matching version of each file")
      parser.add_argument('-d',
         action='store',
         type=parse_date,
         metavar="datetime",
         help="cut-off time (format [[cc]yy]mmdd[-HHMM[SS]]) from which files are not included in "
         "listig or restore. Year defaults to current or previous if date would be in the future. "
         "Time defaults to midnight. Note that to show all files in May, the argument must be 0601")
      parser.add_argument('glob',
         nargs='*',
         help="File path pattern (* matches /) of file to be included in listing or restore")
      return parser.parse_args(argv[1:])

   @property
   def database_path(self):
      return self.archive_root_path+"/index.sl3"

   def backup_path_abs(self, archive_dir_rel, checksum):
      return self.archive_root_path +"/" +archive_dir_rel +"/" +checksum

def parse_date(string, now=None):
   today_str= time.strftime("%Y%m%d", time.localtime(now))
   d_str, minus, t_str = string.partition("-")
   if (t_str=="") == (minus==""):
      if len(d_str) in(4,6,8) and len(t_str) in(0,4,6):
         if d_str.isdigit():
            if t_str=="" or t_str.isdigit():
               full_t_str= t_str + "000000"[len(t_str):]
               full_d_str= today_str[:8-len(d_str)] + d_str
               if len(d_str)==4 and full_d_str > today_str:
                  last_year= int(today_str[:4])-1
                  full_d_str= str(last_year) + d_str
               tm= time.strptime(full_d_str + full_t_str, "%Y%m%d%H%M%S")
               return time.mktime(tm)
   raise ValueError()




import os.path
import queue
import sqlite3
import time
from contextlib import contextmanager

@contextmanager
def open(database_path, allow_create=False):
   db= _Database(database_path, allow_create)
   yield db
   db.close()

class _Database(object):
   def __init__(self, database_path, allow_create):
      self.path= database_path
      self.connection_pool_size= 1
      self.connection_pool= queue.Queue()
      existed= os.path.exists(database_path)
      if not existed and not allow_create:
         raise RuntimeError("could not find database "+database_path)
      for _ in range(self.connection_pool_size):
         conn= sqlite3.connect(database_path, timeout=60, check_same_thread=False)
         conn.isolation_level= None # autocommit
         self.connection_pool.put(conn)
      if not existed:
         self._create_tables()
   def close(self):
      for _ in range(self.connection_pool_size):
         self.connection_pool.get().close()

   @contextmanager
   def _get_connection(self, caller):
      conn= self.connection_pool.get()
      try:
         yield conn
      finally:
         self.connection_pool.put(conn)

   def _make_runname(self, timestamp):
      tm= time.localtime(timestamp)
      # format chosen to make the run name a "word" in most editors:
      return time.strftime("%Y_%m_%d_%H%M", tm)

   def _create_tables(self):
      with self._get_connection("_create_tables") as conn:
         conn.execute("""create table content(
            run_name text not null,
            path text not null,
            timestamp float not null,
            checksum text not null)""")
         conn.execute("create unique index content_pk on content(run_name, path)")
         conn.execute("""create table checksum_cache(
            dev int not null,
            ino int not null,
            timestamp float not null,
            checksum text not null)""")
         conn.execute("create table run(run_name text not null, timestamp float)")
         conn.execute("create unique index run_pk on run(run_name)")
         conn.execute("create unique index checksum_cache_pk on checksum_cache(dev, ino)")
         conn.execute("""create table location(
            checksum text not null,
            archive_dir text not null)""")
         conn.execute("create unique index location_pk on location(checksum)")
         conn.execute("create index location_archivedir on location(archive_dir)")

   def open_backup_run(self):
      run_name= self._make_runname(time.time())
      with self._get_connection("open_backup_run") as conn:
         conn.execute("insert into run(run_name) values(?)", [run_name])
      return run_name

   def complete_backup_run(self, run_name):
      with self._get_connection("close_backup_run") as conn:
         cur= conn.execute("update run set timestamp=? where run_name = ?", [time.time(), run_name])
         if cur.rowcount!=1:
            raise RuntimeError("could not complete run "+run_name)
         conn.execute("delete from run where timestamp is null")

   def get_purgeable_backups(self):
      """Returns iterable of (age_factor, checksum) for all backups that are no longer current.
      File with largest age_factor is most preferred for purging.
      """
      with self._get_connection("get_purgeable_backups") as conn:
         cur= conn.execute("""select CHECKSUM.checksum
                  ,CHECKSUM.birthtime
                  ,min(run.timestamp) as deathtime
               from (select checksum, min(timestamp) as birthtime, max(run_name) as last_alive
                     from content
                     group by checksum
                     ) as CHECKSUM
                  join run on run.run_name > CHECKSUM.last_alive and run.timestamp is not null
                  left join content on content.run_name = run.run_name and content.checksum = CHECKSUM.checksum
               where content.checksum is null
               group by CHECKSUM.checksum, CHECKSUM.birthtime
            """)
         now= time.time()
         return (((now - deathtime) / (deathtime - birthtime), checksum) for checksum,birthtime,deathtime in cur)

   def lookup_checksum(self, st_dev, st_ino, timestamp):
      with self._get_connection("lookup_checksum") as conn:
         for (checksum,) in conn.execute("""
               select checksum
                  from checksum_cache
                  where dev=? and ino=? and timestamp=?""", (st_dev, st_ino, timestamp)):
            return checksum
         return None

   def remember_checksum(self, st_dev, st_ino, timestamp, checksum):
      with self._get_connection("remember_checksum") as conn:
         cur= conn.execute("""
            update checksum_cache
            set timestamp= ?, checksum= ?
            where dev=? and ino=?""", (timestamp, checksum, st_dev, st_ino))
         if cur.rowcount==0:
            conn.execute("insert into checksum_cache(timestamp, checksum, dev, ino) values(?,?,?,?)",
               (timestamp, checksum, st_dev, st_ino))

   def lookup_archivedir(self, checksum):
      with self._get_connection("lookup_archivedir") as conn:
         for (archive_dir,) in conn.execute("select archive_dir from location where checksum=?", [checksum]):
            return archive_dir
         return None

   def remember_archivedir(self, checksum, archive_dir_rel):
      with self._get_connection("remember_archivedir") as conn:
         conn.execute("insert into location(checksum, archive_dir) values(?,?)", (checksum, archive_dir_rel))

   def add_backup_entry(self, run_name, path, timestamp, checksum):
      with self._get_connection("add_backup_entry") as conn:
         conn.execute("""insert into content(run_name, path, timestamp, checksum)
            values(?,?,?,?)""", (run_name, path, timestamp, checksum))

   def getarchivedir_usage(self):
      with self._get_connection("getarchivedir_usage") as conn:
         rows= conn.execute("select archive_dir, count(*) as n from location group by archive_dir")
         return {archive_dir:n for archive_dir,n in rows}

   def get_archive_contents(self, patterns, cutoff_date, show_all):
      """

      :param pattern: like-operator patterns for path to match any of
      :type pattern: string list
      :param cutoff_date: time after which files are ignored or None
      :type cutoff_date: time.time() format
      :param show_all: whether all files should be returned rather than just those from the latest run
      :type show_all: bool
      :return: iterable of (path, timestamp, archive_dir, checksum)
      :rtype:
      """
      with self._get_connection("get_archive_contents") as conn:
         sql= "select distinct path, min(timestamp), archive_dir, content.checksum from content"
         sql_args= []
         where_clauses= [] # to be built now but added later
         where_args= []

         if cutoff_date is None:
            if show_all:
               # show all without cutoff date is just everything
               pass
            else:
               # show one without cutoff means latest run
               where_clauses.append("run_name = (select max(run_name) from content)")
         else:
            if show_all:
               # show all with cutoff date
               where_clauses.append("timestamp <= ?")
               where_args.append(cutoff_date)
            else:
               # show one with cutoff date requires subquery for the latest
               sql += " join (select path as p, max(timestamp) as t from content "
               sql +=       " where timestamp <= ?"
               sql_args.append(cutoff_date)
               sql +=       " group by path) as latest"
               sql +=  " on path = latest.p and timestamp = latest.t"

         sql += " join location on location.checksum = content.checksum"

         if patterns:
            where_clauses.append("("
               +" or ".join("path glob ?" for x in patterns)
               +")")
            where_args.extend(patterns)

         if where_clauses:
            sql += " where " + " and ".join(where_clauses)
            sql_args.extend(where_args)

         sql += " group by path, archive_dir, content.checksum"
         sql += " order by path, min(timestamp) desc"
         yield from conn.execute(sql, sql_args)

   def testing_move_timestamps(self, offset):
      """Adds `offset` to all timestamps in the database, including run names,
      in order to allow backups made at the same time for testing
      to have different names"""
      with self._get_connection("testing_move_timestamps") as conn:
         conn.execute("update run set timestamp= timestamp+?", [offset])
         conn.execute("update content set timestamp= timestamp+?", [offset])

         runs= conn.execute("select run_name, timestamp from run").fetchall()
         params= []
         for old_names, timestamp in runs:
            params.extend((old_names, self._make_runname(timestamp)))
         case_clause= "when ? then ? "*len(runs) +"end"
         conn.execute("update content set run_name= case run_name "+case_clause, params)
         conn.execute("update run set run_name= case run_name "+case_clause, params)
